// Copyright (c) 2021 The Chromium OS Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.
#include <iomanip>
#include "base/logging.h"

#include "arm_spe_decoder.h"

namespace quipper
{

std::string ArmSpePacketDecoder::GetArmParserStateString(ArmParserState state)
{
  if (state == ArmParserState::START) {
    return "START";
  } else if (state == ArmParserState::PC) {
    return "PC";
  } else if (state == ArmParserState::LAT) {
    return "LAT";
  } else if (state == ArmParserState::TYPE) {
    return "TYPE";
  } else if (state == ArmParserState::LD_ST) {
    return "LD_ST";
  } else if (state == ArmParserState::BRANCH) {
    return "BRANCH";
  } else if (state == ArmParserState::TGT) {
    return "TGT";
  } else if (state == ArmParserState::PAD) {
    return "PAD";
  } else if (state == ArmParserState::INST_OTHER) {
    return "INST_OTHER";
  } else if (state == ArmParserState::END) {
    return "END";
  } else if (state == ArmParserState::EXIT) {
    return "EXIT";
  } else if (state == ArmParserState::UNKNOWN) {
    return "UNKNOWN";
  }
  return "UNKNOWN";
}

static std::string Dec2hex(int num)
{
  std::stringstream ioss;
  std::string tmpStr;
  ioss << std::hex << std::setw(2) << std::setfill('0') << num;
  ioss >> tmpStr;
  return tmpStr;
}

void ArmSpePacketDecoder::HandleError()
{
  LOG(ERROR) << "ARM SPE: only the following Arm SPE flags are supported: jitter, branch_filter, load_filter, store_filter, event_filter, min_latency.";
  LOG(ERROR) << "ARM SPE: recommend usage: perf record -e arm_spe_0/jitter=1/ -c COUNT -- COMMAND";
  LOG(ERROR) << "ARM SPE: unsupport arm_spe raw string! error_pos/total_size: " << m_curPos << "/" << m_totalSize;
  LOG(ERROR) << "ARM SPE: current processing status code: " << GetArmParserStateString(m_curParserState);
  std::string errorByteStr;
  for (size_t i = m_curPos; i < m_curPos + 10 && i < m_totalSize; i++) {
    errorByteStr.append(Dec2hex(int(m_rawTraceData[i])));
    errorByteStr.append(" ");
  }
  if (!errorByteStr.empty()) {
    LOG(FATAL) << "ARM SPE: recent error 10 byte: " << errorByteStr;
  } else {
    LOG(FATAL) << "ARM SPE: there is no valid byte.";
  }
}

bool ArmSpePacketDecoder::Init(const std::string* traceData)
{
  if (traceData == nullptr || traceData->empty()) {
    LOG(ERROR) << "TraceData is invalid!";
    return false;
  }
  m_rawTraceData = traceData->c_str();
  m_totalSize = traceData->size();
  m_curParserState = ArmParserState::START;
  m_curPos = 0;
  return true;
}

void ArmSpePacketDecoder::InitProcessArmSpePacketMap()
{
  m_decodeArmSpePacketMap.insert({ArmParserState::START, &ArmSpePacketDecoder::ProcessInitSection});
  m_decodeArmSpePacketMap.insert({ArmParserState::PC, &ArmSpePacketDecoder::ProcessPCSection});
  m_decodeArmSpePacketMap.insert({ArmParserState::LAT, &ArmSpePacketDecoder::ProcessLATSection});
  m_decodeArmSpePacketMap.insert({ArmParserState::TYPE, &ArmSpePacketDecoder::ProcessTypeSection});
  m_decodeArmSpePacketMap.insert({ArmParserState::LD_ST, &ArmSpePacketDecoder::ProcessLDSTSection});
  m_decodeArmSpePacketMap.insert({ArmParserState::BRANCH, &ArmSpePacketDecoder::ProcessBranchTGTSection});
  m_decodeArmSpePacketMap.insert({ArmParserState::INST_OTHER, &ArmSpePacketDecoder::ProcessInstOtherSection});
  m_decodeArmSpePacketMap.insert({ArmParserState::PAD, &ArmSpePacketDecoder::ProcessPadSection});
  m_decodeArmSpePacketMap.insert({ArmParserState::END, &ArmSpePacketDecoder::ProcessEndSection});
}

bool ArmSpePacketDecoder::IsUnsolvedPacketExist() const
{
  if (m_totalSize > 0 && m_curPos < m_totalSize) {
    return true;
  }
  return false;
}

bool ArmSpePacketDecoder::ProcessInitSection(ArmSpeParsedPacket& packet)
{
  if (m_rawTraceData[m_curPos] == ArmSpeInstTag::PC) {
    m_curPos++;
    m_curParserState = ArmParserState::PC;
    return true;
  } else if (m_rawTraceData[m_curPos] == ArmSpeInstTag::PAD) {
    m_curParserState = ArmParserState::PAD;
    return true;
  }
  return false;
}

bool ArmSpePacketDecoder::ProcessPCSection(ArmSpeParsedPacket& packet)
{
  constexpr size_t pcLen = 7;
  memcpy(&packet.sampleAddr, &(m_rawTraceData[m_curPos]), pcLen);
  m_curPos += pcLen;
  m_curPos++; // 1 = en_ls(1)
  m_curParserState = ArmParserState::LAT;
  return true;
}

bool ArmSpePacketDecoder::ProcessLATSection(ArmSpeParsedPacket& packet)
{
  /* LAT is useless for AutoFDO, ignore it. */
  m_curPos += 6; // 6 = LAT(6)
  m_curParserState = ArmParserState::TYPE;
  return true;
}

bool ArmSpePacketDecoder::ProcessLDSTSection(ArmSpeParsedPacket& packet)
{
  /* Virtual Address and LAT is useless for AutoFDO, ignore it. */
  m_curPos += 15; // 15 = VA(9) + LAT(6)
  m_curParserState = ArmParserState::PAD;
  m_curInstType = ArmSpeInstType::ARM_SPE_INST_LD_ST;
  return true;
}

bool ArmSpePacketDecoder::ProcessBranchTGTSection(ArmSpeParsedPacket& packet)
{
  if (m_rawTraceData[m_curPos] == ArmSpeBrTypeTag::B_COND) {
    m_curInstType = ArmSpeInstType::ARM_SPE_INST_BR_COND;
  } else if (m_rawTraceData[m_curPos] == ArmSpeBrTypeTag::B_IND) {
    m_curInstType = ArmSpeInstType::ARM_SPE_INST_BR_IND;
  } else if (m_rawTraceData[m_curPos] == ArmSpeBrTypeTag::B) {
    m_curInstType = ArmSpeInstType::ARM_SPE_INST_BR;
  } else {
    LOG(ERROR) << "ARM SPE: Unsupport branch tag!";
    return false;
  }
  m_curPos += 2; // 2 = BRANCH_TYPE(1) + TGT_TAG(1)
  constexpr size_t pcLen = 7; // 7 = PC(7)
  memcpy(&packet.targetAddr, &(m_rawTraceData[m_curPos]), pcLen);
  m_curPos += pcLen + 1; // 1 = el_ns(1)
  m_curParserState = ArmParserState::PAD;
  return true;
}

bool ArmSpePacketDecoder::ProcessInstOtherSection(ArmSpeParsedPacket& packet)
{
  m_curInstType = ArmSpeInstType::ARM_SPE_INST_OTHER;
  m_curParserState = ArmParserState::PAD;
  return true;
}

bool ArmSpePacketDecoder::ProcessPadSection(ArmSpeParsedPacket& packet)
{
  while(m_rawTraceData[m_curPos] == ArmSpeInstTag::PAD && m_curPos < m_totalSize - 1) {
    m_curPos++;
  }
  m_curParserState = ArmParserState::END;
  return true;
}

bool ArmSpePacketDecoder::ProcessEndSection(ArmSpeParsedPacket& packet)
{
  m_curParserState = ArmParserState::EXIT;
  if (m_rawTraceData[m_curPos] == ArmSpeInstTag::PAD || m_rawTraceData[m_curPos] == ArmSpeInstTag::END) {
    /* PAD tag means this is the end of Arm SPE section, there is no END tag.
       END tag means that there are still packet s to be processed next. */
    m_curPos++;
    return true;
  } else if (m_rawTraceData[m_curPos] == ArmSpeInstTag::PC) {
    /* sometimes packet not end with ArmSpeTag::END but with
       ArmSpeTag::PC, so it should be processed next. */
    return true;
  }
  LOG(ERROR) << "ARM SPE: Invalid end section!";
  return false;
}

bool ArmSpePacketDecoder::ProcessTypeSection(ArmSpeParsedPacket& packet)
{
  if (m_rawTraceData[m_curPos++] != ArmSpeInstTag::TYPE) {
    LOG(ERROR) << "ARM SPE: Unsupport sample type section tag!";
    return false;
  }

  memcpy(&packet.packetType, &(m_rawTraceData[m_curPos]), sizeof(packet.packetType));
  m_curPos += sizeof(packet.packetType);

  if (m_rawTraceData[m_curPos] == ArmSpeInstTag::LD_ST) {
    m_curParserState = ArmParserState::LD_ST;
    m_curPos += 2; // 2 = FLAG(2)
  } else if (m_rawTraceData[m_curPos] == ArmSpeInstTag::BRANCH) {
    m_curParserState = ArmParserState::BRANCH;
    m_curPos++; // 1 = BR_TAG(1)
  } else if (m_rawTraceData[m_curPos] == ArmSpeInstTag::INST_OTHER) {
    m_curParserState = ArmParserState::INST_OTHER;
    m_curPos += 2; // 2 = FLAG(2)
  } else {
    LOG(ERROR) << "ARM SPE: Unsupport instruction type tag!";
    return false;
  }
  return true;
}

bool ArmSpePacketDecoder::SolveOnePacket(ArmSpeParsedPacket& packet)
{
  m_curParserState = ArmParserState::START;
  while (m_curParserState != ArmParserState::EXIT) {
    if (!(this->*m_decodeArmSpePacketMap[m_curParserState])(packet)) {
      return false;
    }
  }
  return true;
}

} // namespace quipper
