// Copyright (c) 2021 The Chromium OS Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.
#ifndef PERF_DATA_CONVENTER_ARM_SPE_PARSER
#define PERF_DATA_CONVENTER_ARM_SPE_PARSER

#include <vector>
#include <list>

#include "arm_spe_decoder.h"

namespace quipper {
/*
  This class is used to parser Arm SPE events,
  usage as flollows: SetTraceDataPtr() -> ParserRawSpeData() -> GetParsedSpeEvent()
*/

class ArmSpeParser {
public:
  explicit ArmSpeParser() {}
  ~ArmSpeParser() {}

  /* Parsing raw Arm SPE byte stream, return the result. */
  bool ParsingArmSpeData(const std::string* traceStrPtr) {
    if (!SetTraceDatePtr(traceStrPtr)) {
      return false;
    }
    return ParserRawSpeData();
  }

  /* Return a ref of parsed SPE packets. */
  const std::list<ArmSpeParsedPacket>& GetParsedSpeEvent() const {
    return m_parsedSpeEvents;
  }

  /* ArmSpeEventType is a highly compressed format, we need decode it into strings when use. */
  const std::vector<std::string> GetArmSpeEventNameString(ArmSpeEventType) const;

  /* Free saving parsedSpeEvents. */
  void Clear() {
    m_parsedSpeEvents.clear();
  }

  bool ArmSpeTraceDataExist() const {
    return m_armSpeTraceDataExist;
  }

  void SetArmSpeTraceDataExist(bool exist) {
    m_armSpeTraceDataExist = exist;
  }

private:
  /* traceDataPtr is a pointer to Arm SPE data string, it must be called firstly.
     DO NOT FREE traceStrPtr here, it`s belong to other modules. */
  bool SetTraceDatePtr(const std::string* traceStrPtr) {
    if (traceStrPtr == nullptr || traceStrPtr->empty()) {
      return false;
    }
    m_rawTraceData = traceStrPtr;

    return true;
  }

  /* Parsing raw Arm SPE data, the target event will be saved in m_parsedSpeEvents. */
  bool ParserRawSpeData();

  /* Return true if there is any target event we want. */
  bool IsTargetEvent(const ArmSpeParsedPacket&) const;

  inline void SavePacket(const ArmSpeParsedPacket& packet) {
    m_parsedSpeEvents.push_back(packet);
  }

  /* Record whether Arm SPE sample exists from PERF_RECORD_AUXTRACE_INFO. */
  bool m_armSpeTraceDataExist = false;

  /* Pointer to raw Arm SPE data, DO NOT FREE in this class. */
  const std::string* m_rawTraceData = nullptr;

  /* Save parsed Arm Spe event. */
  std::list<ArmSpeParsedPacket> m_parsedSpeEvents;

  /* Decoder is used for parser raw Arm SPE packet. */
  ArmSpePacketDecoder m_armSpePacketDecoder;

}; // class ArmSpeParser

} // namespace quipper

#endif // define PERF_DATA_CONVENTER_ARM_SPE_PARSER